
# This file was generated by 'versioneer.py' (0.14) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

version_version = '0.47.0'
version_full = '4eb9cf875969ef7f099b292e0c5a8bb852f6fc7f'
def get_versions(default={}, verbose=False):
    return {'version': version_version, 'full': version_full}

